#!/usr/bin/env python

import sys

if sys.version_info<(3,0,0):
	import pygtk as PyGtk
	import gtk   as Gtk
	import gtk   as GLib
else:
	from gi.repository import Gtk
	from gi.repository import GLib

if sys.version_info<(3,0,0):
	ICON_SIZE_BUTTON = Gtk.ICON_SIZE_BUTTON
else:
	ICON_SIZE_BUTTON = Gtk.IconSize.BUTTON

class Visual:
	FRAMES_PER_SEC = 60 # hz
	RATES_PER_SEC  = 1  # hz

	def __init__(self, config, device, logger):
		def get_objects(names):
			return list(map(lambda x:
				self.builder.get_object(x), names))

		self.config   = config
		self.device   = device
		self.logger   = logger
		self.builder  = Gtk.Builder()
		self.history  = [0.5]*1000

		self.rate_due = 0
		self.rate_cnt = {'acc': 0,
		                 'mag': 0,
		                 'lgt': 0,
		                 'tch': 0,
		                 'a2d': 0}

		self.builder.add_from_file('visual.ui')
		self.builder.connect_signals(self)

		self.window   = self.builder.get_object("window")
		self.settings = self.builder.get_object("settings")
		self.plotdraw = self.builder.get_object('plot')
		self.plotsns  = self.builder.get_object('sns')
		self.plotaxis = self.builder.get_object('axis')

		if sys.version_info<(3,0,0):
			self.plotdraw.connect('expose-event', self.on_expose)
		else:
			self.plotdraw.connect('draw', self.on_draw)

		self.accs     = get_objects(['accx', 'accy', 'accz'])
		self.mags     = get_objects(['magx', 'magy', 'magz'])
		self.lgts     = get_objects(['lgt'])
		self.tchs     = get_objects(['tchp', 'tchd'])
		self.a2ds     = get_objects(['a2d0', 'a2d1', 'a2d2', 'a2d3', 'a2d4', 'a2d5'])

	# Signal handlers
	def on_hide(self, win, *args):
		self.settings.hide()
		return True

	def on_show(self, act):
		self.settings.show()
		return True

	def on_serial(self, act):
		if act.get_active():
			self.status(self.device.connect())
		else:
			self.device.disconnect()
		return True

	def on_xively(self, act):
		if act.get_active():
			self.status(self.logger.connect())
		else:
			self.logger.disconnect()
		return True

	def on_flush(self, act):
		self.logger.flush()

	def on_enable(self, obj):
		name  = Gtk.Buildable.get_name(obj)
		name  = name.replace('_btn', '')
		state = obj.get_active()
		if self.config.enable[name] != state:
			self.device.set_enable(name, state)
			self.config.enable[name] = state
			self.config.save()

	def on_rate(self, obj):
		name  = Gtk.Buildable.get_name(obj)
		name  = name.replace('_spin', '')
		value = obj.get_value()
		if self.config.rate[name] != value:
			self.device.set_rate(name, value)
			self.config.rate[name] = value
			self.config.save()

	def on_config(self, obj):
		name = Gtk.Buildable.get_name(obj)
		if isinstance(obj, Gtk.SpinButton):
			value = obj.get_value()
		else:
			value = obj.get_text()
		if getattr(self.config, name) != value:
			setattr(self.config, name, value)
			self.config.save()

	def on_draw(self, obj, cairo):
		count   = len(self.history)
		extents = cairo.clip_extents()
		width   = extents[2] - extents[0]
		height  = extents[3] - extents[1]

		cairo.set_line_width(4)
		for i in range(count):
			x = i*width/(count-1)
			y = ((1-self.history[i]) * 0.8 + 0.1) * height
			if i==0:
				cairo.move_to(x, y)
			else:
				cairo.line_to(x, y)
		cairo.stroke()

	def on_expose(self, obj, event):
		cairo = obj.window.cairo_create()
		self.draw(obj, cairo)

	def on_key(self, win, ev):
		if ev.string == 'q':
			Gtk.main_quit();

	def on_destroy(self, win):
		Gtk.main_quit()

	# Methods
	def load_config(self):
		def set_val(name, val):
			self.builder.get_object(name).set_value(val)
		def set_txt(name, val):
			self.builder.get_object(name).set_text(val)
		def set_act(name, val):
			self.builder.get_object(name).set_active(val)

		# Serial settings
		set_txt('device',   self.config.device)
		set_val('baudrate', self.config.baudrate)
		set_txt('parity',   self.config.parity)
		set_val('databits', self.config.databits)
		set_val('stopbits', self.config.stopbits)

		# Xively settings
		set_txt('feedid',   self.config.feedid)
		set_txt('apikey',   self.config.apikey)
		set_val('maxrate',  self.config.maxrate)

		# Sensor settings
		for key in list(self.config.enable.keys()):
			set_act(key+'_btn',  self.config.enable[key])
			set_val(key+'_spin', self.config.rate[key])

	def update(self, state):
		def setval(objs, vals, lbls):
			found = 0
			for i in range(0,len(objs)):
				if vals[i] == None:
					continue
				found = 1
				if lbls and lbls[i]:
					objs[i].set_text('%s: %f' % (lbls[i], vals[i]))
				else:
					objs[i].set_text('%f' % vals[i])
			return found

		# Dispaly values and sum up rates
		self.rate_cnt['acc'] += setval(self.accs, state.acc, ['X', 'Y', 'Z'])
		self.rate_cnt['mag'] += setval(self.mags, state.mag, ['X', 'Y', 'Z'])
		self.rate_cnt['lgt'] += setval(self.lgts, state.lgt, [])
		self.rate_cnt['tch'] += setval(self.tchs, state.tch, ['P', 'D'])
		self.rate_cnt['a2d'] += setval(self.a2ds, state.a2d, [])

		# Save history
		idxs = ['acc', 'mag', 'tch', 'lgt', 'a2d']
		sns  = idxs[self.plotsns.get_active()]
		axis = int(self.plotaxis.get_value())
		if hasattr(state, sns):
			data = getattr(state, sns)
			axis = min(axis-1, len(data)-1)
			item = data[axis]
			if item != None:
				self.history = self.history[1:] + [item]

	def debug(self, state):
		print('update: ' + str(state))
		print('\tacc - ' + str(state.acc))
		print('\tmag - ' + str(state.mag))
		print('\tlgt - ' + str(state.lgt))
		print('\ttch - ' + str(state.touch))
		print('\ta2d - ' + str(state.a2d))

	def read_timer(self):
		def check(state, name):
			btn = self.builder.get_object(name + '_btn')
			con = self.builder.get_object(name + '_con')
			img = Gtk.STOCK_YES if state else Gtk.STOCK_NO
			btn.set_active(state)
			con.set_from_stock(img, ICON_SIZE_BUTTON)

		# Update status icons
		check(self.device.running(), "serial")
		check(self.logger.running(), "xively")

		# Read data and update data display
		for item in self.device.process():
			self.update(item)
			self.logger.append(item)
			#self.debug(item)

		# Refresh window
		self.window.queue_draw()

		return True

	def rate_timer(self):
		for sns in self.rate_cnt:
			rate  = self.rate_cnt[sns] * Visual.RATES_PER_SEC
			obj   = self.builder.get_object(sns + '_rate')
			obj.set_text('%d' % rate)
			self.rate_cnt[sns] = 0
		return True

	def run(self):
		self.load_config()
		self.window.show()
		GLib.timeout_add(1000/Visual.FRAMES_PER_SEC, self.read_timer)
		GLib.timeout_add(1000/Visual.RATES_PER_SEC,  self.rate_timer)
		Gtk.main()

	# Private methods
	def status(self, msg):
		status = self.builder.get_object("status")
		status.pop(context_id=0)
		if not msg:
			return
		status.push(text=('Error: ' + msg), context_id=0)
